;;; nnimap.el --- IMAP backend for Gnus

;;; Copyright (C) 1998 Simon Josefsson <jas@pdc.kth.se>
;;; Copyright (C) 1998 Jim Radford <radford@robby.caltech.edu>
;;; Copyright (C) 1997 John McClary Prevost <visigoth@cs.cmu.edu>

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

;;; Commentary:

;;; This file provides IMAP support for the Emacs news/mailreader Gnus.
;;; To enable this backend you put the following in your .gnus:

;;;  (require  'nnimap)
;;;  (setq gnus-secondary-select-methods 
;;;      '((nnimap 
;;;          "yoyo"
;;;          (nnimap-server-address "robby.caltech.edu"))))

;;; And inside Gnus, list all available groups with A A (search for 
;;; 'nnimap') and subscribe to the mailboxes you are interested in with U. 
;;; If you know the name of the mailbox, you can also use 'U'
;;; (gnus-group-unsubscribe-group) to subscribe to it (no mailbox
;;; completion here, sorry).

;;; Note that Gnus can't deal with groups starting with '.' (a dot),
;;; this means that you will have to use ~/iMail/ instead of ./iMail
;;; (or similair).

;;; Note also that you shouldn't subscribe to goups on your nnimap
;;; server from the server buffer if the server is your primary server
;;; i.e. in `gnus-select-method'.  I think this is a bug in Gnus.
;;; Yep, I just checked.  It is fixed in at least 5.6.27.  It was
;;; still broken in 5.5.

;;; RFC1730 (IMAP4): done
;;; RFC1731 (Authentication mecanisms): currently only support for KERBEROS_V4
;;; RFC2060 (IMAP4rev1): done
;;; RFC2086 (ACL's): done
;;; RFC2195 (CRAM-MD5): done
;;; RFC2359 (UIDPLUS ext): done

;;; Todo (roughly in order of priority):

;;;   o Convert the manual to texi.
;;;   o Verify that we don't use IMAP4rev1 specific things (RFC2060 App B)
;;;   o nnimap-request-close-group should do incremental flag setting
;;;     instead of first clearing all flags and setting them again. It also
;;;     shouldn't try to set flags at all in read-only groups.
;;;   o split to other backends, different split rules for different
;;;     servers/inboxes
;;;   o fix nnimap-request-list 'fast and default to it if possible
;;;   o Do we dare to send the EXPUNGE command?
;;;   o Split up big fetches (1,* header especially) in smaller chunks
;;;   o use \Draft to support the draft group??
;;;   o What do I do with gnus-newsgroup-*?
;;;   o Figure out when update-info is called and why and if we need to
;;;     call it more often and from where.
;;;   o Add update-info calls? To request-group, close-group, retrieve-groups
;;;   o Tell Gnus about new groups (how can we tell?)
;;;   o Add asynchronous support with two server connections (use BODY.PEEK)
;;;   o Bulletin board stuff.
;;;   o Respooling (fix Gnus?)
;;;   o Subscribe to newly created mailboxes (how? nnimap-request-newgroups?)
;;;   o Add support for the following: (if applicable)

;;;       request-list-newsgroups, request-regenerate
;;;       list-active-group, request-update-mark,
;;;       request-expire-articles,
;;;       request-associate-buffer, request-restore-buffer,

;;;   o When UIDVALIDITY changed, I should reconstruct everything gnus knows
;;;     about the group (possible?)
;;;   o UTF-7 encode mailbox names (does Gnus deal with SPC in group names?)
;;;   o Fix Gnus to handle leading '.' in group names
;;;   o Fix the flag situation when using the Gnus Agent
;;;   o Support RFC2221 (Login referrals)
;;;   o IMAP2BIS compatibility (RFC2061)
;;;   o Debug imtest, it dumps with "Time is out of bounds" sometimes
;;;   o ACAP stuff (perhaps a different project, would be nice to ACAPify 
;;;     .newsrc.eld)
;;;   o MIME
;;;   

(require 'imap)

(require 'nnoo)
(require 'nnheader)
(require 'nnmail)
(require 'gnus)
(require 'gnus-range)

(eval-when-compile (require 'cl))

(gnus-declare-backend "nnimap" 'mail 'address 'prompt-address 
		      'physical-address 'respool)  ;; respool??

(nnoo-declare nnimap) ; we derive from no one

(defconst nnimap-version "nnimap 0.3.26")

;; External variables.

(defvoo nnimap-list-pattern "*" 
"*PATTERN or list of PATTERNS use to limit available groups.  

A pattern is either GROUP or (REFERENCE . GROUP).

GROUP is a string.  See the available wildcard characters below.

The meaning of REFERENCE is server-specific, so it's
expected that you (the user) can figure out the appropriate setting.
On the Cyrus server, this is irrelevant.  On the UWash server, this
gets joined together with GROUP.  If it is not specified the
default is an empty string.

Example:
 '(\"INBOX\" \"Mail/*\" \"alt.sex.*\" (\"~friend/Mail/\" . \"list/*\"))

Also note that currently groups that start with a '.' cause Gnus
to choke (a patch is pending acceptance), so instead of something
like \"*\" which might match \".mailboxlist\" you could use
\"~/*\" which would match \"~/.mailboxlist\" and not cause problems.

The two wildcards are * and %.  * means match anything, much like in
shell globbing in Unix.  * does match hierarchy delimiters (. or /, in
the usual case.)  % is the same as *, but does not match the hierarchy
symbol.")

(defvoo nnimap-list-method "LIST" ; "LSUB"
  "*Function called on IMAP server to list groups.  One of \"LSUB\" or
\"LIST\". LSUB means only retrieve groups marked on the server as
subscribed.  LIST means every matching group should be retrieved.")

(defvoo nnimap-server-address nil
  "*The name of the IMAP server.  If nil, uses the virtual server's name.

The port should be specified with nnimap-server-port, but can also be
done here by using the port-extended syntax:

server.name:143

Of course, the default port is 143 so you would specify some other
port if you need to use this.")

(defvoo nnimap-server-port nil
  "*The port of the IMAP server.  If nil, uses the default port. (143).

If you have multiple servers on the same host (but different ports),
you can't use this option together with nnimap-authinfo, since netrc
is a one-service file. You should then specify the port in
nnimap-server-address using the port-extended syntax, and use the same
port-extended address in your authinfo file, and don't use this
variable at all.")

;; Splitting variables

(defvar nnimap-split-crosspost t
  "If non-nil, do crossposting if several split methods match the mail.
If nil, the first match found will be used.")

(defvar nnimap-split-inbox nil
  "*Name of mailbox to split mail from. Mail is read from this mailbox and
split according to rules in nnimap-split-rules.

This can be a string or a list of strings.")

(defvar nnimap-split-rule nil
  "*Mail will be split according to theese rules. Mail is read from mailbox(es)
specified in nnimap-split-inbox.

If you'd like, for instance, one mail group for mail from the
\"gnus-imap\" mailing list, one group for junk mail and leave
everything else in the incoming mailbox, you could do something like
this:

(setq nnimap-split-rule '((\"INBOX.gnus-imap\"   \"From:.*gnus-imap\")
			  (\"INBOX.junk\"        \"Subject:.*buy\")))

As you can see, nnimap-split-rule is a list of lists, where the first
element in each \"rule\" is the name of the IMAP mailbox, and the
second is a regexp that nnimap will try to match on the header to find
a fit.

The first element can also be a list. In that case, the first element
is the server the second element is the group on that server in which
the matching article will be stored.

The second element can also be a function.  In that case, it will be
called narrowed to the headers with the first element of the rule as
the argument.  It should return a non-nil value if it thinks that the
mail belongs in that group.")

(defvoo nnimap-imap-defs nil
  "*Definitions of variables to set up in the IMAP buffer.")

(defvoo nnimap-group-list-speed 'slow ; 'fast, 'medium
  "*If fast, do not show number of articles in the group list.
If medium, guess number of articles by using the UIDNEXT attribute.
If slow, fetch the UID of lowest/highest article.")

;; Authorization / Privacy variables

;; todo:
;; better 'smart that open a connection, check the capabilities and
;; set this variable to the "best" method. easy, but we should fallback
;; if it doesn't work.
(defvoo nnimap-auth-method 'smart
  "How nnimap authenticate itself to the server.

The default, `smart', will try to use the \"best\" method the server
can handle.

Unfortunely, the `smart' option isn't all that clever and will not
recognize KERBEROS_V4 or SSL so you should set this if you use them.

Available options: smart, login, cram-md5, kerberos4 and ssl.")

(defcustom nnimap-authinfo-file "~/.authinfo"
  "Authorization information for IMAP servers. In .netrc format."
  :type
  '(choice file
	   (repeat :tag "Entries"
		   :menu-tag "Inline"
		   (list :format "%v"
			 :value ("" ("login" . "") ("password" . ""))
			 (string :tag "Host")
			 (checklist :inline t
				    (cons :format "%v"
					  (const :format "" "login")
					  (string :format "Login: %v"))
				    (cons :format "%v"
					  (const :format "" "password")
					  (string :format "Password: %v")))))))



;; Internal variables.

(defvoo nnimap-group-alist nil)
(defvoo nnimap-server-buffer nil)

;; radford:
;;  shouldn't have to keep this info around, nnoo should???
;;  The only reason for this is `nnimap-server-opened' needs
;;  `nnimap-server-buffer' to do its work, but have to change
;;  servers to get the right one.  I don't think `opened'
;;  should be changing servers.  A better way anyone???
;; jas:
;;  i think this is the way to do it. from "Backend interface" in
;;  the gnus manual:
;;  "When Gnus asks for information from a backend--say nntp---on
;;  something, it will normally include a virtual server name in the
;;  function parameters. (If not, the backend should use the "current"
;;  virtual server.) For instance, nntp-request-list takes a virtual
;;  server as its only (optional) parameter. If this virtual server hasn't
;;  been opened, the function should fail."
(defvar nnimap-server-buffer-alist nil)

(defvar nnimap-debug "*nnimap-debug*")

(when nnimap-debug
  (require 'trace)
  (buffer-disable-undo (get-buffer-create nnimap-debug))
  (mapc (lambda (f) (trace-function-background f nnimap-debug))
        '(
          nnimap-close-group
          nnimap-close-server
          nnimap-open-server  
	  nnimap-authenticate
          nnimap-possibly-change-group
          nnimap-possibly-change-server
	  nnimap-request-accept-article
          nnimap-request-article
          nnimap-request-article-part
          nnimap-request-body
          nnimap-request-close
          nnimap-request-create-group
          nnimap-request-delete-group
          nnimap-request-group
          nnimap-request-head
          nnimap-request-list
          nnimap-request-list-mapper
          nnimap-request-newgroups
          nnimap-request-post
          nnimap-request-rename-group
          nnimap-request-scan
;	  nnimap-send-command-wait
;	  nnimap-ok-p
	  nnimap-split-copy-delete-article
          nnimap-split-move-article
          nnimap-split-to-groups
	  nnimap-split-articles
          nnimap-request-update-info
          nnimap-request-update-mark
          nnimap-retrieve-groups
          nnimap-retrieve-headers
          nnimap-server-opened
          nnimap-status-message
	  nnimap-update-alist-soft
	  nnimap-search
	  nnimap-range-to-string
          nnimap-request-expire-articles
          nnimap-request-move-article
	  nnimap-edit-acl-done
	  nnimap-edit-acl
          )))


;;; Interface functions, required backend functions

;;; This seems to define the following:
;;;  `nnimap-close-server'       `nnimap-server-opened'
;;;  `nnimap-open-server'        `nnimap-status-message'
;;; open-server -> nnoo-change-server
;;; close-server -> nnoo-close-server
;;; server-opened -> nnoo-server-opened
;;; status-message -> nnoo-status-message

(nnoo-define-basics nnimap)

;; todo:
;; handle "MESSAGE" too
(defun nnimap-body-lines (body)
  (if (listp body)
      (if (and (stringp (car body))
	       (string= (car body) "TEXT")
	       (numberp (nth 7 body)))
	  (nth 7 body)
      (apply '+ (mapcar 'nnimap-body-lines body)))
    0))

;; todo:
;; use NOV lines instead? A fetch like
;;   (UID RFC822.SIZE BODY BODY[HEADER.FIELDS (References)]) would do it
;; remove redundant lines:/chars:
(deffoo nnimap-retrieve-headers (articles &optional group server fetch-old)
  (when (nnimap-possibly-change-group group server)
    (with-current-buffer nntp-server-buffer (erase-buffer))
    (let* (compressed (uncompressed (gnus-uncompress-sequence articles)))
      (when (numberp fetch-old) 
        (setcar uncompressed (- (car uncompressed) fetch-old)))
      (setq compressed (gnus-compress-sequence uncompressed t))
      ;; Should we clear `imap-data-folder' ???
      (with-current-buffer nnimap-server-buffer
        (when 
            ;; Should we clear `imap-message-data' before this???  If
            ;; we don't need to cache the headers then I think we
            ;; should.
            (nnimap-ok-p (nnimap-send-command-wait
                          (concat "UID FETCH "
                                  (if (and fetch-old (not (numberp fetch-old)))
                                      "1:*"
                                    (nnimap-range-to-string compressed))
                                  " (UID RFC822.HEADER RFC822.SIZE BODY)")))
          (mapc (lambda (num)
                  (let* ((header (imap-message-get num 'RFC822.HEADER))
			 (size   (imap-message-get num 'RFC822.SIZE))
			 (body   (imap-message-get num 'BODY))
			 (lines  (nnimap-body-lines body)))
                    (with-current-buffer nntp-server-buffer
                      (if (not header)
                          (insert "423 Bad article number.\n")
                        (insert (format "221 %d Article retrieved.\n" num))
			(insert (format "Chars: %d\n" size))
			(insert (format "Lines: %d\n" lines))
                        (insert header)
                        (delete-char -1)
                        (insert ".\n")))))
                uncompressed)
          'headers)))))

(defun nnimap-authenticate (server &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    ;; Login
    (let (list alist user passwd ok)
      (and (fboundp 'gnus-parse-netrc)
	   (setq list (gnus-parse-netrc nnimap-authinfo-file)
		 alist (gnus-netrc-machine
			list nnimap-server-address)
		 imap-username (gnus-netrc-get alist "login")
		 imap-password (gnus-netrc-get alist "password")))
      (cond ((eq nnimap-auth-method 'smart)
	     (and (memq 'AUTH=CRAM-MD5 (imap-capability-get))
		  (setq imap-auth-method 'imap-authenticate-cram-md5)))
	    ((eq nnimap-auth-method 'md5) 
	     (setq imap-auth-method 'imap-authenticate-cram-md5)))
      (setq ok (imap-authenticate server buffer))
      ;; only loop if 1) the server didn't validate us AND 2) the user
      ;; actually typed something
      (while (and (not ok)
		  imap-username
		  (not (not imap-password))) ;; no password in backtraces
	(message "Bad Password for %s, Try again." server)
	(sleep-for 2)
	(setq ok (imap-authenticate server buffer)))
      ok)))
      
(deffoo nnimap-open-server (server &optional defs)
  (or (and (nnimap-server-opened server)
	   (nnoo-change-server 'nnimap server defs)
           (if (null nnimap-server-buffer)
               (error "this shouldn't happen")
             t))
      (progn
        (unless (assq 'nnimap-server-address defs)
          (push (list 'nnimap-server-address server) defs))
        (unless (assq 'nnimap-server-buffer defs)
          (push (list 'nnimap-server-buffer (concat " *nnimap* " server)) defs))
	(nnoo-change-server 'nnimap server defs)
	(let* ((pos (string-match ":" nnimap-server-address))
	       (port (and pos 
			  (string-to-number 
			   (substring nnimap-server-address (1+ pos)))))
	       (host (substring nnimap-server-address 0 pos)))
	  (condition-case err
	      ;; todo: do clever things here if nnimap-auth-method is 'smart
	      (let ((imap-open-stream (cond ((eq nnimap-auth-method 
						 'kerberos4)
					     'imap-open-imtest-stream)
					    ((eq nnimap-auth-method
						 'ssl)
					     'imap-open-ssl-stream))))
		(when (imap-open-server host (or port nnimap-server-port)
					nnimap-server-buffer 
					nnimap-imap-defs)
		  ;; we only support imap4.*
		  (unless (and (nnimap-ok-p (nnimap-send-command-wait 
					     "CAPABILITY" nnimap-server-buffer))
			       (or (memq 'IMAP4 (imap-capability-get
						 nnimap-server-buffer))
				   (memq 'IMAP4rev1 (imap-capability-get
						     nnimap-server-buffer))
				   (memq 'IMAP4REV1 (imap-capability-get
						     nnimap-server-buffer))))
		    (imap-close-server nnimap-server-buffer)
		    (nnheader-report 
		     'nnimap 
		     "Sorry, %s is not a IMAP4(rev1) server." host)
		    (sit-for 2)
		    ;; Make sure we close the server
		    (signal 'quit))
		  ;; authenticate ourself
		  (when (nnimap-authenticate server nnimap-server-buffer)
		    (push (cons server nnimap-server-buffer)
			  nnimap-server-buffer-alist))))
	    ;; Assume error comes from host unknow so we are not connected yet
	    (error (kill-buffer nnimap-server-buffer)
		   (nnheader-report 'nnimap (cdr err)))
	    (quit (let (imap-last-status)
		    (imap-close-server nnimap-server-buffer)
		    (unless (get-buffer nnimap-server-buffer)
		      (kill-buffer nnimap-server-buffer))
		    (nnheader-report 'nnimap (cdr err)))))))))
  
(deffoo nnimap-close-server (&optional server)
  (let ((s-b (assoc 
              (setq server (or server (nnoo-current-server 'nnimap)))
              nnimap-server-buffer-alist)))
    (when s-b
      (setq nnimap-server-buffer nil)
      (setq nnimap-server-buffer-alist (delq s-b nnimap-server-buffer-alist))
      (imap-close-server (cdr s-b))
      (kill-buffer (cdr s-b)))
    (nnoo-close-server 'nnimap server)))

(deffoo nnimap-request-close ()
  (mapc (lambda (s-b) (nnimap-close-server (car s-b)))
        nnimap-server-buffer-alist)
  (setq nnimap-server-buffer-alist nil))

(deffoo nnimap-server-opened (&optional server)
  (let ((buffer (cdr (assoc 
                      (setq server (or server (nnoo-current-server 'nnimap)))
                      nnimap-server-buffer-alist))))
    (if (and (gnus-buffer-live-p buffer)
	     (gnus-buffer-live-p nntp-server-buffer))
	(let ((running (imap-server-opened buffer)))
	  ;; clean up autologouts
	  (unless running
	    (nnimap-close-server server))
	  running))))

;  (and (nnoo-server-opened 'nnimap server)
;       (imap-server-opened nnimap-server-buffer)))

(deffoo nnimap-status-message (&optional server)
  (let ((buffer (cdr (assoc 
                      (setq server (or server (nnoo-current-server 'nnimap)))
                      nnimap-server-buffer-alist))))
    (when buffer
      (with-current-buffer buffer
	(or (cdr imap-last-status)
	    (nnoo-status-message 'nnimap server))))))

(deffoo nnimap-request-article (article &optional group server to-buffer)
  (nnimap-request-article-part 'RFC822 article group server to-buffer t))
(deffoo nnimap-request-head    (article &optional group server to-buffer)
  (nnimap-request-article-part 'RFC822.HEADER article group server to-buffer))
(deffoo nnimap-request-body    (article &optional group server to-buffer)
  (nnimap-request-article-part 'RFC822.TEXT article group server to-buffer t))

(defun nnimap-request-article-part (part article &optional 
					 group server to-buffer peek)
  (when (nnimap-possibly-change-group group server)
    (with-current-buffer nnimap-server-buffer
      (when (stringp article)
        ;; Find the article number by Message-Id
        (let ((articles (nnimap-search 
			 (format "HEADER Message-Id %d" article))))
          (when (cdr articles)
            (message 
             "IMAP Warning: Message Id search yielded more than one article %s"
             articles))
          (setq article (car articles))))
      (when article
        ;; Find the article by number
        (nnimap-send-command-wait (format "UID FETCH %d (%s%s)" article part
					  (if peek ".PEEK" "")))
        (let ((text (imap-message-get article part)))
          (with-current-buffer (or to-buffer nntp-server-buffer)
            (erase-buffer)
            (if (not text)
                (nnheader-report 
		 'nnimap 
		 (format "Article %s does not exist." article))
              (insert text) 
	      t)))))))

;;; Select GROUP and unless FAST return 211 EXISTS LOWEST HIGHEST GROUP
;;; The active info returned goes into the `gnus-active-hashtb'.
;;; Don't call this directly, call `gnus-activate-group'.
;;; `gnus-activate-group' can call `request-scan' before request-group.
;;; Ok, here seems to be the problem.  We return 'group and
;;; `gnus-groups-to-gnus-format' seem to not use a prefix "nnimap+yoyo:"
;;; when entering something into the active-hashtb but it does when
;;; it looks for new news.  Damn.
;;;
;;; Also they seem to want us to update the active-info, but
;;; we can;t give a good answer for the lowest message number,
;;; so I think that request-update-info should somehow override
;;; this once it figures out what the real lowest is.
;;;
;;; The 'fast' concept is broken. Gnus =relies= on getting data
;;; back from this function even if called with fast=t. Yuck.
(deffoo nnimap-request-group (group &optional server fast)
  (nnimap-request-update-info group (gnus-get-info 
				     (gnus-group-normally-qualified 
				      'nnimap server group))
			      server
			      t)
  (when (nnimap-possibly-change-group group server) ;; SELECT group
;    (if fast
;	(with-current-buffer nntp-server-buffer 
;	  (erase-buffer)
;	  t)
      (with-current-buffer nnimap-server-buffer
	(let ((imap-message-data (make-vector 2 0))
	      (exists (imap-folder-get 'EXISTS))
	      articles)
	  (if (eq 0 exists)
	      (setq articles '(0))
	    (when (nnimap-ok-p (nnimap-send-command-wait "FETCH 1,* (UID)"))
	      (imap-message-map (lambda (uid Uid)
				  (push uid articles)) 'UID)
	      ;; start of bug workaround code
	      ;; Lotus CCmail is broken, returns:
	      ;; C13 FETCH 1,* (UID)
	      ;; * 1 FETCH (UID 4198)
	      ;; C13 OK Completed
	      ;; we fetch * in addition
	      (when (< 2 (length articles))
		(when (nnimap-ok-p (nnimap-send-command-wait "FETCH * (UID"))
		  (imap-message-map (lambda (uid Uid)
				      (push uid articles)) 'UID)))
	      ;; end of bug workaround code
	      ))
	  (when articles
	    (with-current-buffer nntp-server-buffer
	      (erase-buffer)
	      (insert (format "211 %d %d %d %s\n" 
			      exists (max 1 (apply 'min articles))
			      (apply 'max articles) group)))
	    t)))));)

;; Note that there is no need for this in current Gnus (5.6.27), all
;; you need to do is use gnus-group-prefixed-name.  I'm not sure when
;; this got fixed.  That's what I get for not using the current
;; version.
(defun gnus-group-normally-qualified (backend server group)
  ;; This is the format for groups in the group-info.
  ;; native groups are not qualified.  Why?
  (let ((method (gnus-server-to-method (format "%s:%s" backend server))))
    (if (gnus-method-equal gnus-select-method method)
        ;; Native (we of course can't call `gnus-native-p' since it
        ;; assumes group is the result of this function
        group
      (gnus-group-prefixed-name group method))))
            
(defvar nnimap-mark-to-flag-alist
  (mapcar 
   (lambda (pair) ; cdr is the mark
     (or (assoc (cdr pair)
                '((read . "\\Seen")
                  (tick . "\\Flagged")
                  ;;(expire . "\\Deleted")
                  (draft . "\\Draft")
                  (reply . "\\Answered")))
         (cons (cdr pair)
               (format "gnus-%s" (symbol-name (cdr pair))))))
   (cons '(read . read) gnus-article-mark-lists)))

;;; We need to optimize for no changes i.e. doing Q.  We should
;;; not touch the server.
(deffoo nnimap-close-group (group &optional server)
  (when (nnimap-possibly-change-group group server)
    (setq server (or server (nnoo-current-server 'nnimap)))
    (with-current-buffer nnimap-server-buffer
      ;; For now we assume that we need to sync the group-info
      ;; with the server here.
      (if (and (imap-folder-get 'writable)
	       (imap-folder-get 'permanentflags group))
      (let* ((info (gnus-get-info 
                    (gnus-group-normally-qualified 'nnimap server group)))
                 (marks (gnus-info-marks info)))
	(unless (eq 0 (imap-folder-get 'EXISTS))
	  ;; Cyrus server (v1.5.2) disconnects on empty groups ???
	  (nnimap-send-command-wait 
               (concat "UID STORE 1:* -FLAGS.SILENT ("
                       (mapconcat 
                        'identity
                        (remove "\\*" (imap-folder-get 'permanentflags group))
                        " ")
                       ")")))
        (mapc
             (lambda (mark-flag)
               (if (nnimap-flag-permanent-p mark-flag group)
                   (let ((val (if (equal 'read (car mark-flag))
                                  (gnus-info-read info)
                                (cdr (assoc (car mark-flag) marks)))))
                     (if val
               (nnimap-send-command-wait 
                (concat "UID STORE " 
                                  (nnimap-range-to-string val)
                                  " +FLAGS.SILENT (" (cdr mark-flag) ")"))))))
             nnimap-mark-to-flag-alist)))
      ;; When someone Qs out of a group we could EXAMINE first???
      ;; Close her up.  We don't necessarily have to do this.
      (when (or (and (memq 'UNSELECT (imap-capability-get))
		     (nnimap-ok-p (nnimap-send-command-wait "UNSELECT")))
		(and (nnimap-ok-p 
		      (nnimap-send-command-wait (concat "EXAMINE " group)))
		     (nnimap-ok-p (nnimap-send-command-wait "CLOSE"))))
        (setq imap-current-folder nil
              imap-message-data nil)
	t))))

;;; Returns: GROUP HIGHEST LOWEST [ynmxj=]
(defun nnimap-request-list-mapper (group)
  (unless (or (member '\\NoSelect (imap-folder-get 'flags group))
	      ;; We ignore groups with spaces (Gnus can't handle them)
	      (string-match " " group))
    (let (high)
      (gnus-message 7 "Generating active list for nnimap group %s" group)
      (cond 
       ((eq nnimap-group-list-speed 'slow)
	(when (nnimap-possibly-change-group group nil)
	  (let ((imap-message-data (make-vector 2 0))
		(exists (imap-folder-get 'EXISTS))
		articles)
	    (if (eq 0 exists)
		(with-current-buffer nntp-server-buffer
		  (insert (format "%s 0 1 y\n" group)))
	      (when (nnimap-ok-p (nnimap-send-command-wait
				  "FETCH 1,* (UID)"))
		(imap-message-map (lambda (uid Uid) 
				    (push uid articles)) 'UID)
		;; start of bug workaround code
		;; Lotus CCmail is broken, returns:
		;; C13 FETCH 1,* (UID)
		;; * 1 FETCH (UID 4198)
		;; C13 OK Completed
		;; we fetch * in addition
		(when (< 2 (length articles))
		  (when (nnimap-ok-p (nnimap-send-command-wait 
				      "FETCH * (UID"))
		    (imap-message-map (lambda (uid Uid)
					(push uid articles)) 'UID)))
		;; end of bug workaround code
		(when articles
		  (with-current-buffer nntp-server-buffer
		    (insert (format "%s %d %d y\n" group
				    (apply 'max articles)
				    (apply 'min articles)))
		    t)))))))
       ((eq nnimap-group-list-speed 'medium)
	;; Need to quote group if it has spaces!!!???
	(when (nnimap-ok-p (nnimap-send-command-wait
			    (concat "STATUS " group 
				    " (UIDNEXT)")))
	  (setq high (imap-folder-get 'UIDNEXT group))
	  (with-current-buffer nntp-server-buffer
	    (insert (format "%s %d 1 y\n" group (- high 1)))
	    t)))
       ((eq nnimap-group-list-speed 'fast)
	(with-current-buffer nntp-server-buffer
	  (insert (format "%s 0 1 y\n" group))
	  t))
       (t
	(error "Unknown nnimap-group-list-speed: %s"
	       nnimap-group-list-speed))))))

;;; Returns: GROUP HIGHEST LOWEST [ynmxj=]
(deffoo nnimap-request-list (&optional server)
  (when (nnimap-possibly-change-server server)
    (with-current-buffer nntp-server-buffer (erase-buffer))
    (with-current-buffer nnimap-server-buffer
      (setq imap-data-folder (make-vector imap-convenient-group-prime 0)
	    imap-current-folder nil)
      (message "ok")
      (let* ((patterns (mapcar (lambda (p) (if (consp p)
					       p 
					     (cons "\"\"" p)))
			       (if (and (listp nnimap-list-pattern)
					(listp (cdr nnimap-list-pattern)))
				   nnimap-list-pattern
				 (list nnimap-list-pattern))))
	     pattern)
	(while (setq pattern (pop patterns))
	  (nnimap-ok-p (nnimap-send-command-wait
                        (concat nnimap-list-method " "
                                (car pattern) " "
                                (cdr pattern)))))
        (gnus-message 5 "Generating active list for %s" server)
	(let ((nnimap-group-list-speed 'fast))
	  (mapatoms (lambda (sym) (nnimap-request-list-mapper 
				   (symbol-name sym)))
		    imap-data-folder)))
    (setq nnimap-group-alist (nnmail-get-active))
    t)))
  

;;; IMAP doesn't support posting, but this must be defined
(deffoo nnimap-request-post (&optional server)
  nil)


;;; Interface functions, optional backend functions

;;; This gets called instead of `nnimap-request-list' when
;;; `gnus-read-active-file' is 'some instead of t.
;;; Returns: GROUP HIGHEST LOWEST [ynmxj=] ???
(deffoo nnimap-retrieve-groups (groups &optional server)
  (when (nnimap-possibly-change-server server)
    (with-current-buffer nntp-server-buffer (erase-buffer))
    (with-current-buffer nnimap-server-buffer
      ;; Force the slow method for now since this will only be called
      ;; for subscribed groups. 
      (let (group)
        (gnus-message 5 "Generating active info for certain groups on %s"
                      server)
        (while (setq group (pop groups))
          (nnimap-request-list-mapper group)))
      ;; (setq nnimap-group-alist (nnmail-get-active)))
      'active)))

(defvar nnimap-mark-to-predicate-alist
  (mapcar 
   (lambda (pair) ; cdr is the mark
     (or (assoc (cdr pair)
                '(;;(read . "SEEN")
                  (tick . "FLAGGED")
                  ;;(expire . "DELETED")
                  (draft . "DRAFT")
                  (reply . "ANSWERED")))
         (cons (cdr pair)
               (format "KEYWORD gnus-%s" (symbol-name (cdr pair))))))
   gnus-article-mark-lists))

(defun nnimap-flag-permanent-p (mark-string group)
  (or (member (cdr (assoc (car mark-string) 
                          nnimap-mark-to-flag-alist))
	      (imap-folder-get 'permanentflags group))
      (and (string-match "gnus-" (cdr mark-string))
           (member "\\*" (imap-folder-get 'permanentflags group)))))
  
;;; I really think this should update the active-info too???
(deffoo nnimap-request-update-info (group info &optional server silly)
  (when silly
    (when (nnimap-possibly-change-group group server) ;; SELECT
      (with-current-buffer nnimap-server-buffer
	(gnus-message 5 "Updating active info for mailbox %s" group)
	
	;; extend info to have parameters (otherwise when we set them, 
	;; they don't get set)
	(or (gnus-info-params info) 
	    (gnus-info-set-params info nil t))
	
	;; Replace list of read and marked articles with authoritative
	;; data from server
	(gnus-info-set-read 
	 info 
	 ;; xxx This is extremely slow.
	 (let* (
		;; oldseen could contain articles marked unread by other
		;; imap clients!  we correct this
		(oldseentmp (gnus-uncompress-range (gnus-info-read info)))
		(unseen (nnimap-search "UNSEEN"))
		(oldseen (gnus-set-difference oldseentmp unseen))
		;; oldseen might lack articles marked as read by other
		;; imap clients! we correct this
		(newseen (nnimap-search "SEEN"))
		;; ok, read articles are in oldseen+newseen (xxx lots of dupes)
		(seen (append oldseen newseen))
		;; sort to have gnus-compress-sequence remove dupes
		(seens (sort seen '<))
		(read (gnus-compress-sequence seens t)))
	   (if (and read
		    (atom (car read)))
	       (list (cons (car read) (car read))) ;; xxx not my bug
	     read)))
	(mapc
	 (lambda (mark-search) 
	   (if (nnimap-flag-permanent-p mark-search group)
	       (gnus-info-set-marks 
		info 
		(nnimap-update-alist-soft (car mark-search) 
					  (gnus-compress-sequence 
					   (nnimap-search (cdr mark-search)))
					  (gnus-info-marks info)))))
	 nnimap-mark-to-predicate-alist))
      info)))

;;; Respond to articles with mail
(deffoo nnimap-request-type (group article)
  'mail)

;; Move article ARTICLE from group GROUP on current server to group
;; TO-GROUP.  
(defun nnimap-split-copy-delete-article (article group to-group)
  (when (nnimap-ok-p (nnimap-send-command-wait
		      (format "UID COPY %d %s" article to-group)))
    (let ((store 
	   (nnimap-send-command-wait 
	    (format "UID STORE %d +FLAGS (\\Seen \\Deleted)" article))))
      (if (nnimap-ok-p store)
	  (message "IMAP split: moved %s:%d to %s" group article to-group)
	(error "IMAP flag store failed: %s (you may have unread mail marked as read!)" store)))))

(defun nnimap-split-move-article (article group to-group)
  (when to-group
    (unless (nnimap-split-copy-delete-article article group to-group)
      (message "Could not find mailbox %s, creating..." to-group)
      (if (nnimap-ok-p (nnimap-send-command-wait
			(format "CREATE %s" to-group)))
	  (nnimap-split-copy-delete-article article group to-group)
	(message "Could not create mailbox %s: %s" 
		 to-group imap-last-status)))))

;; doesn't use COPY but instead FETCH+APPEND. Slower.
(defun nnimap-split--move-article (article group to-group)
  (when to-group
    (message "IMAP split: moving %s:%d to %s" group article to-group)
    
    ;; mailbox exist?  create it or bail out
    (with-current-buffer nnimap-server-buffer (setq imap-data-folder [0]))
    (when (and (nnimap-ok-p (nnimap-send-command-wait 
			     (concat "LIST \"\" " to-group)
			     nnimap-server-buffer))
	       (if (imap-folder-get 'delimiter to-group 
				    nnimap-server-buffer)
		   t
		 (if (nnimap-ok-p (nnimap-send-command-wait
				   (concat "CREATE " to-group)
				   nnimap-server-buffer))
		     t
		   (progn 
		     (message "Couldn't create group %s, leaving article in %s"
			      to-group group)
		     nil))))
      ;; ok, group exist, move article
      (unless (nnimap-request-move-article 
	       article group nil
	       (list 'nnimap-request-accept-article to-group))
	(message "Couldn't move article %s:%d to %s" group article to-group)
	;; remove Seen flag on article -- request-move-article fetches
	;; the whole article, which set's this flag. Bad.
	(unless (nnimap-ok-p 
		 (nnimap-send-command-wait 
		  (format "UID STORE %d -FLAGS.SILENT \\Seen" article)))
	  ;; serious damage
	  (error "Couldn't remove Seen flag on article %s:%d (you might have unread mail flagged as read!)" group article))))))

;; tries to match all rules in nnimap-split-rule against content of 
;; nntp-server-buffer, returns a list of groups that matched.
(defun nnimap-split-to-groups (rule)
  (let (to-groups)
    (with-current-buffer nntp-server-buffer
      (mapcar (lambda (rule) 
		(let ((group (car rule))
		      (regexp (cadr rule)))
		  (goto-char (point-min))
		  (when (re-search-forward regexp nil t)
		    (setq to-groups (cons group to-groups)))))
	      rule)
      (reverse to-groups))))

(defun nnimap-split-find-rule (server inbox)
  nnimap-split-rule)

(defun nnimap-split-find-inbox (server)
  (if (listp nnimap-split-inbox)
      nnimap-split-inbox
    (list nnimap-split-inbox)))

;; 47
(defun nnimap-split-articles (&optional group server)
  (when (nnimap-possibly-change-server server)
    (with-current-buffer nnimap-server-buffer
      (let (rule inbox (inboxes (nnimap-split-find-inbox server)))
	;; iterate over inboxes
	(while (and (setq inbox (pop inboxes))
		    (nnimap-possibly-change-group inbox)) ;; SELECT
	  ;; find split rule for this server / inbox
	  (when (setq rule (nnimap-split-find-rule server inbox))
	    (let (article (unseens (nnimap-search "UNSEEN")))
	      ;; iterate over articles
	      (while (setq article (pop unseens))
		(when (nnimap-request-head article)
		  ;; article into what groups?
		  (let ((groups (nnimap-split-to-groups rule)))
		    ;; move it there
		    (if nnimap-split-crosspost
			;; move to all boxes
			(let (to-group)
			  (while (setq to-group (pop groups))
			    (nnimap-split-move-article article inbox 
						       to-group)))
		      ;; move to first matching box, if any
		      (nnimap-split-move-article article inbox 
						 (car groups))))))))
	      (when (nnimap-possibly-change-group inbox) ;; just in case
		;; todo: only remove splitted articles
		;; use UID EXPUNGE, unfortunely not in the standard :-(
		(nnimap-send-command-wait "EXPUNGE") ;; to get tagged data
		(when (nnimap-send-command-wait "CLOSE")
		  (setq imap-current-folder nil
			imap-message-data nil))))))
    t))

;; nnimap-request-scan doesn't need to do anything. the delivery agent
;; program is responsible for putting new message in the imap folders.
;; compare the situation with nntp (nil function) and with all nnmail-
;; backends (fetches mail from spools, POPing, saving in files on local
;; storage etc).
;; On the other hand, we could do message splitting here.
(deffoo nnimap-request-scan (&optional group server)
  (nnimap-split-articles group server))

;; until everyone uses gnus >= 5.6.24
(deffoo nnimap-request-group-description (group &optional server)
  (when (nnimap-possibly-change-server server)
    (with-current-buffer nntp-server-buffer
      (erase-buffer))
    t))
 
;; until everyone uses gnus >= 5.6.24
(deffoo nnimap-request-list-newsgroups (&optional server)
  (when (nnimap-possibly-change-server server)
    (with-current-buffer nntp-server-buffer
      (erase-buffer))
    t))

;; until everyone uses gnus >= 5.6.24
(deffoo nnimap-request-newgroups (date &optional server)
  (when (nnimap-possibly-change-server server)
    (with-current-buffer nntp-server-buffer
      (erase-buffer))
    t))
    
;; (deffoo nnimap-request-update-mark (group article mark)

;;; Maybe we should add to current group-alist???
(deffoo nnimap-request-create-group (group &optional server args)
  (when (nnimap-possibly-change-server server)
    (nnimap-ok-p (nnimap-send-command-wait
                  (concat "CREATE " group) 
                  nnimap-server-buffer))))

;;; Notice that we don't actually delete anything, we just mark them
;;; deleted.  We need some kind of EXPUNGE command.  Also, when we
;;; expunge we delete all marked \Deleted articles.  So we either
;;; need to unmark,expunge,re-mark or come up with some new scheme.
;;; I guess force means delete, not force means expire.  We don't to
;;; any expiry currently.
;;; Also, we can search for articles older than some expiry time if
;;; we need to.  I'm not sure what to do.
(deffoo nnimap-request-expire-articles (articles group &optional server force)
  (when force
    (when (nnimap-possibly-change-group group server)
      (with-current-buffer nnimap-server-buffer
;;;        (setq deleted (nnimap-search "DELETED"))
;;;        (nnimap-ok-p (nnimap-send-command-wait "EXPUNGE"))
;;;        (setq articles (gnus-sorted-complement articles deleted)))))
        (when (nnimap-ok-p (nnimap-send-command-wait 
                            (concat "UID STORE "
                                    (nnimap-range-to-string
                                     (gnus-compress-sequence articles))
                                    " +FLAGS (\\Deleted)")))
          (setq articles nil)))))
  ;; return articles not deleted
  articles)

(deffoo nnimap-request-move-article (article group server
					     accept-form &optional last)
  (save-excursion
    (let ((buf (get-buffer-create " *nnimap move*"))
	  result)
      (and 
       (nnimap-request-article article group server)
       (save-excursion
	 (set-buffer buf)
	 (buffer-disable-undo (current-buffer))
	 (insert-buffer-substring nntp-server-buffer)
	 (setq result (eval accept-form))
	 (kill-buffer buf)
	 result)
       (nnimap-request-expire-articles (list article) group server t))
      result)))

;;; todo:
;;; Maybe we should update the group-alist???
(deffoo nnimap-request-accept-article (group &optional server last)
  (when (nnimap-possibly-change-server server)
    ;; The article will be appended as UIDNEXT
    (when (or (memq 'UIDPLUS (imap-capability-get nnimap-server-buffer))
	      (nnimap-ok-p (nnimap-send-command-wait
			    (concat "STATUS " group " (UIDNEXT)")
			    nnimap-server-buffer)))
      (with-current-buffer (current-buffer)
	(goto-char (point-min))
	(unless (string= "\n" imap-eol)
	  (while (re-search-forward "\n" nil t)
	    (replace-match imap-eol))))
      (when (nnimap-ok-p (nnimap-send-command-wait 
			  ;; Optional flags,date???
			  (list (concat "APPEND " group " ") 
				(current-buffer))
			  nnimap-server-buffer))
	(let ((high (cdr (imap-folder-get 
			  (if (memq 'UIDPLUS (imap-capability-get 
					      nnimap-server-buffer))
			      'appenduid
			    'UIDNEXT)
			  group nnimap-server-buffer))))
	  (when high
	    (cons group high)))))))

;; (deffoo nnimap-request-replace-article -- IMAP does not support replacing

;;; Maybe we should remove from group-alist and unset if current group???
(deffoo nnimap-request-delete-group (group force &optional server)
  (when (nnimap-possibly-change-server server)
    (nnimap-ok-p (nnimap-send-command-wait
                  (concat "DELETE " group)
                  nnimap-server-buffer))))

;;; Maybe we should update the group-alist and current group???
(deffoo nnimap-request-rename-group (group new-name &optional server)
  (when (nnimap-possibly-change-server server)
    (when (string= "INBOX" group)
      (nnheader-report 
       'nnimap 
       "Renaming INBOX will only move content to a newly created mailbox")
      (sit-for 1))
    (nnimap-ok-p (nnimap-send-command-wait
                  (concat "RENAME " group " " new-name)
                  nnimap-server-buffer))))


;;; Internal functions

(defun nnimap-update-alist-soft (key value alist)
  (if value 
      (cons (cons key value) (remassoc key alist))
    (remassoc key alist)))

(defun nnimap-range-to-string (range)
  (mapconcat 
   (lambda (item)
     (if (consp item)
         (format "%d:%d" 
                 (car item) (cdr item))
       (format "%d" item)))
   (if (and (listp range) (not (listp (cdr range))))
       (list range) ;; make (1 . 2) into ((1 . 2))
     range)
   ","))

(defun nnimap-send-command-wait (command &optional buffer)
  (unless (and (stringp command) (string-match "%" command))
    (gnus-message 10 (apply 'concat
			    (if (listp command)
				(mapcar
				 (lambda (s-b) 
				   (cond 
				    ((bufferp s-b) (buffer-name s-b))
				    ((functionp s-b) "<func>")
				    (s-b)))
				 command)
			      (list command)))))
  (imap-send-command-wait command buffer))

(defun nnimap-ok-p (status)
  (if status
      (if (imap-ok-p status)
	  t
	(nnheader-report 'nnimap (cdr status)))
    (nnheader-report 'nnimap (format "IMAP Command Timed Out"))))

(defun nnimap-search (predicate)
  (imap-folder-set 'search nil)
  (when (nnimap-ok-p (nnimap-send-command-wait 
		      (concat "UID SEARCH " predicate)))
    (imap-folder-get 'search)))

(defun nnimap-possibly-change-server (server)
  "Change to server SERVER if needed (open it if it's closed). If SERVER is
nil, change to current server."
  (and (setq server (or server (nnoo-current-server 'nnimap)))
       (nnimap-open-server server)))

(defun nnimap-possibly-change-group (group &optional server)
  (when (nnimap-possibly-change-server server)
    (setq server (or server (nnoo-current-server 'nnimap)))
    (with-current-buffer nnimap-server-buffer
      (when group
	(unless (string= group imap-current-folder)
	  (setq imap-message-data (make-vector 199 0)) ; itty bitty obarray???
	  ;; Folder must be set during SELECT
	  (setq imap-current-folder group)
          (let ((groupname (gnus-group-normally-qualified
                            'nnimap server group)))
            (if (not (nnimap-ok-p (nnimap-send-command-wait
                                   (concat (if (gnus-ephemeral-group-p
                                                groupname)
                                               "EXAMINE"
                                             "SELECT")
                                           " " group))))
                ;; Failed SELECT unselects the current group
                (setq imap-current-folder nil)
              ;; check/set UIDVALIDITY
              (let ((new-uid (imap-folder-get 'uidvalidity))
                    (old-uid (gnus-group-get-parameter groupname 
						       'uidvalidity))
		    (info (gnus-get-info groupname)))
                (if (not old-uid)
		    (gnus-group-add-parameter groupname (cons 'uidvalidity 
							      new-uid))
		  (when (not (equal old-uid new-uid))
		    (if (and (not (gnus-info-marks info))
			     (not (gnus-info-read info)))
			(gnus-group-set-parameter groupname 'uidvalidity 
						  new-uid)
                      (message "UIDVALIDITY clash. Old value `%s', new `%s'"
                               old-uid new-uid)
                      (setq imap-current-folder nil)))))))))
      imap-current-folder)))

;;; Other functions

(defun gnus-group-edit-nnimap-acl (group)
  "Edit the Access Control List of current nnimap GROUP."
  (interactive (list (gnus-group-group-name)))
  (let ((mailbox (gnus-group-real-name group))
	method acl)
    (unless group
      (error "No group on current line"))
    (unless (gnus-get-info group)
      (error "Killed group; can't be edited"))
    (unless (eq 'nnimap (car (setq method (gnus-find-method-for-group group))))
      (error "ACL editing only available for nnimap groups"))
    (when (nnimap-possibly-change-server (cadr method))
      (unless (memq 'ACL (imap-capability-get nnimap-server-buffer))
	(error "Your server does not support ACL editing"))
      (gnus-edit-form (with-current-buffer nnimap-server-buffer
			(imap-folder-set 'acl nil mailbox)
			(nnimap-send-command-wait (format "GETACL %s" mailbox))
			(let ((raw (imap-folder-get 'acl mailbox))
			      user access)
			  (while (setq access (pop raw))
			    (if user
				(prog1
				    (push (cons (symbol-name user)
						(symbol-name access)) acl)
				  (setq user nil))
			      (setq user access))))
			acl)
		      (format "Editing the access control list for `%s'.

   An access control list is a list of (identifier . rights) elements.

   The identifier specifies the corresponding user. The identifier
   `anyone' is reserved to refer to the universal identity.

   Rights is a string listing a (possibly empty) set of alphanumeric
   characters, each character listing a set of operations which is being
   controlled. Letters are reserved for ``standard'' rights, listed
   below.  Digits are reserved for implementation or site defined rights.

   l - lookup (mailbox is visible to LIST/LSUB commands)
   r - read (SELECT the mailbox, perform CHECK, FETCH, PARTIAL,
       SEARCH, COPY from mailbox)
   s - keep seen/unseen information across sessions (STORE SEEN flag)
   w - write (STORE flags other than SEEN and DELETED)
   i - insert (perform APPEND, COPY into mailbox)
   p - post (send mail to submission address for mailbox,
       not enforced by IMAP4 itself)
   c - create (CREATE new sub-mailboxes in any implementation-defined
       hierarchy)
   d - delete (STORE DELETED flag, perform EXPUNGE)
   a - administer (perform SETACL)" group)
		      `(lambda (form) 
			 (gnus-group-edit-nnimap-acl-done 
			  ,mailbox ',method ',acl form))))))

(defun gnus-group-edit-nnimap-acl-done (mailbox method old-acls acls)
  (when (nnimap-possibly-change-server (cadr method))
    (with-current-buffer nnimap-server-buffer
      ;; delete all removed identifiers
      (let ((deleted (copy-list old-acls))
	    (dontdelete acls) acl)
	(while (setq acl (pop deleted))
	  (unless (assoc (car acl) dontdelete)
	    (let ((status 
		   (nnimap-send-command-wait
		    (format "DELETEACL %s %s" mailbox (car acl)))))
	      (when (and (listp status)
			 (eq 'NO (car status)))
		(error "Can't delete ACL: %s" (cadr status))))))
      ;; set all changed acl's
      (let ((new-acls acls) acl)
	(while (setq acl (pop new-acls))
	  (let* ((user (car acl))
		 (access (cdr acl))
		 (old-access (cdr (assoc user old-acls))))
	    (unless (string= access old-access)
	      (let ((status 
		     (nnimap-send-command-wait
		      (format "SETACL %s %s %s" mailbox user access))))
		(when (and (listp status)
			   (eq 'NO (car status)))
		  (error "Can't set ACL: %s" (cadr status))))))))))))

;;; Glue

(defun nnimap-group-mode-hook ()
  (define-key gnus-group-mode-map (if (fboundp 'kbd) (kbd "G l")
				    (read-kbd-macro "G l"))
    'gnus-group-edit-nnimap-acl))
(add-hook 'gnus-group-mode-hook 'nnimap-group-mode-hook)

;; We're done.

(provide 'nnimap)
